#ifndef __VHEX_DISPLAY_FONT__
# define __VHEX_DISPLAY_FONT__

#include <vhex/defs/types.h>
#include <vhex/defs/attributes.h>
#include <vhex/display/shader.h>

/* font_t: Font data encoded for dfont*() functions */
typedef struct font {
	/* Font name (NUL-terminated), NULL if no title */
	char const *name;

	/* Font shape flags */
	byte_union(shape,
		uint8_t bold	:1;
		uint8_t italic	:1;
		uint8_t serif	:1;
		uint8_t mono	:1;
		uint8_t 	:3;
		uint8_t prop	:1;
	);

	/* Character spacing (usually 1) */
	uint8_t char_spacing;

	/* glyph information */
	struct {
		uint8_t height;
		uint8_t line_height;
		uint32_t *data;
		uint32_t count;

		union {
			/* For monospaced fonts */
			struct {
				uint8_t width;
				uint16_t storage_size;
			} mono;
			/* For proportional fonts */
			struct __workaround {
				uint8_t width;
				uint16_t index;
				uint8_t shift;
			} *prop;
		};
	} glyph;

	/* unicode blocks */
	struct {
		struct {
			/* Unicode point of first character in block */
			uint32_t start   :20;
			/* Length of block */
			uint32_t length  :12;
		} *blocks;
		uint8_t block_count;
	} unicode;

} VPACKED(4) font_t;

/* dfont_get() : get the current font */
extern font_t *dfont_get(void);

/* dfont_text_geometry() : get the char geometry */
extern int dfont_text_geometry(
	font_t *font,
	char const * const str_char,
	int *size,
	size_t *w,
	size_t *h
);


//---
// Kernel-level API
//---

/* dfont_render() : draw caracter in surface */
extern void dfont_text_render(struct dshader_surface *surface, uint32_t *arg);

#endif /* __VHEX_DISPLAY_FONT__ */
