#include <vhex/display/font.h>

//---
// Internal
//---

/* dfont_utf8_next(): Read the next UTF-8 code point of a string */
static uint32_t dfont_utf8_next(uint8_t const **str_pointer)
{
	uint8_t const *str = *str_pointer;
	uint8_t lead = *str++;

	/* Skip non-leaders which are invalid as starting bytes */
	while((lead >= 0x80 && lead <= 0xbf) ||
		lead == 0xc0 || lead == 0xc1 || lead == 0xfe || lead == 0xff) {
		lead = *str++;
	}

	/* This base case will handle the NUL terminator */
	if(lead <= 0x7f) {
		*str_pointer = str;
		return lead;
	}

	uint8_t n2 = (*str++ & 0x3f);
	if(lead <= 0xdf) {
		*str_pointer = str;
		return ((lead & 0x1f) << 6) | n2;
	}

	uint8_t n3 = (*str++ & 0x3f);
	if(lead <= 0xef) {
		*str_pointer = str;
		return ((lead & 0x0f) << 12) | (n2 << 6) | n3;
	}

	uint8_t n4 = (*str++ & 0x3f);
	if(lead <= 0xf7) {
		*str_pointer = str;
		return ((lead & 0x07) << 18) | (n2 << 12) | (n3 << 6) | n4;
	}

	/* It the string is too invalid, force a space and try to continue */
	*str_pointer = str;
	return 0x20;
}

#if 0
/* dfont_glyph_index(): Obtain the glyph index of a Unicode code point */
int dfont_glyph_index(font_t const *f, uint32_t code_point)
{
	int glyph_start = 0;

	for(int i = 0; i < f->block_count; i++) {
		int diff = code_point - f->blocks[i].start;
		if(diff >= 0 && diff < f->blocks[i].length) {
			return glyph_start + diff;
		}

		glyph_start += f->blocks[i].length;
	}

	return -1;
}
#endif

/* dfont_glyph_index(): Obtain the glyph index of a Unicode code point */
//TODO: suport unicode block
static int dfont_glyph_index(font_t const *f, uint32_t code_point)
{
	code_point -= ' ';
	if ((int)code_point < 0 || code_point >= f->glyph.count)
		return (-1);
	return code_point;
}

/* dfont_char_render() : */
static void dfont_char_render(
	struct dshader_surface *surface,
	int glyph_idx,
	int x,
	int y
) {
	int glyph_shift;
	int glyph_width;
	int glyph_size;
//	int surfa_idx;
//	int surfa_shift;
	font_t *font;
	int counter;
	int sx;

	(void)surface;

	/* generate font index / shift information */
	font = dfont_get();
	if (font->shape.prop == 1) {
		glyph_width = font->glyph.prop[glyph_idx].width;
		glyph_shift = font->glyph.prop[glyph_idx].shift;
		glyph_idx   = font->glyph.prop[glyph_idx].index;
		glyph_size  = glyph_width * font->glyph.height;
	} else {
		glyph_width = font->glyph.mono.width;
		glyph_size  = font->glyph.mono.storage_size;
		glyph_shift = glyph_size * glyph_idx;
		glyph_idx   = glyph_shift >> 8;
		glyph_shift = glyph_shift & 0xff;
	}

	/* drawing algo */
	//TODO: cache surface data (limit read/write)
	sx = x;
	counter = 0;
	glyph_shift = 0x80000000 >> glyph_shift;
	for (int i = 0; i < glyph_size; ++i)
	{
#if 0
		if (x >= surface->x
				&& x <= surfa_mx
				&& y >= surface->y
				&& y <= surfa_my) {
			tmp0 = (y - surface->y) * 396;
			tmp1 = (x - surface->x);
			surfa_idx   = (tmp0 + tmp1) >> 2;
			surfa_shift = (tmp0 + tmp1) & 0xff;
			if (font->glyph.data[glyph_idx] & glyph_shift) {
				surface->frag[surfa_idx] = (uint16_t)arg[4];
			} else {
				surface->frag[surfa_idx] = (uint16_t)arg[5];
			}
		}
#endif

		/* update font bitmap index / shift */
		glyph_shift >>= 1;
		if (glyph_shift == 0x00000000) {
			glyph_shift = 0x80000000;
			glyph_idx += 1;
		}

		/* update bitmap position */
		x += 1;
		counter += 1;
		if (counter >= glyph_width) {
			counter = 0;
			x = sx;
			y += 1;
		}
	}
}

//---
// Kernel API
//---


/* dfont_render() : draw caracter in surface */
void dfont_text_render(struct dshader_surface *surface, uint32_t *arg)
{
	int x;
	int y;
	int mx;
	int my;

	/* check culling */
	x = arg[0];
	y = arg[1];
	mx = arg[2];
	my = arg[3];
	if (my < surface->y
			|| mx < surface->x
			|| y > (int)(surface->y + surface->height)
			|| x > (int)(surface->x + surface->width)) {
		return;
	}

	uint8_t const *str = (void*)(uintptr_t)arg[6];
	int nb_char = arg[7];
	int counter = -1;
	uint32_t code_point;
	font_t *font = dfont_get();
	int glyph_idx;
	int glyph_width;

	/* draw algo (update me) */
	while (++counter < nb_char)
	{
		/* get glyph geometry information */
		code_point = dfont_utf8_next(&str);
		glyph_idx = dfont_glyph_index(font, code_point);
		glyph_width = font->glyph.mono.width;
		if (font->shape.prop == 1)
			glyph_width = font->glyph.prop[glyph_idx].width;

		/* check char culling */
		if (x >= surface->x
				&& y >= surface->y
				&& y < (int)(surface->y + surface->height)
				&& x < (int)(surface->x + surface->width)) {
			dfont_char_render(surface, glyph_idx, x, y);
		}

		/* line discipline */
		x += glyph_width + font->char_spacing;
		if (code_point == '\n') {
			y += font->glyph.height;
			x = arg[0];
		}
	}
}

//---
// Public API
//---


/* dfont_get() : get the current font */
font_t *dfont_get(void)
{
	extern font_t font8x9;
	return &font8x9;
}

/* dfont_geometry() : get the char geometry */
int dfont_text_geometry(
	font_t *font,
	char const * const str_char,
	int *size,
	size_t *w,
	size_t *h
) {
	uint8_t const *str0 = (void *)str_char;
	uint8_t const *str1 = (void *)str_char;
	uint32_t code_point;
	size_t char_width;
	size_t length = 0;
	size_t mx = 0;
	size_t x = 0;
	size_t y = 0;
	int glyphidx;
	int limit;

	/* handle special behaviour */
	if (font == NULL)
		font = dfont_get();
	limit = -1;
	if (size != NULL && *size >= 0)
		limit = *size;

	/* generate geometry information */
	while (1) {
		code_point = dfont_utf8_next(&str0);
		if (code_point == 0 || (limit > 0 && str0 - str1 >= limit))
			break;

		char_width = font->glyph.mono.width;
		if (font->shape.prop == 1) {
			glyphidx = dfont_glyph_index(font, code_point);
			if (glyphidx < 0)
				return (-1);
			char_width = font->glyph.prop[glyphidx].width;
		}

		/* line discipline */
		x += char_width + font->char_spacing;
		if (code_point == '\n') {
			y += font->glyph.height;
			x = 0;
		}
		if (x > mx) mx = x;

		length++;
	}

	/* set geometry information */
	if (w != NULL)    *w = mx - font->char_spacing;
	if (h != NULL)    *h = y;
	if (size != NULL) *size = length;
	return (0);
}
